/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2020-2021 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

/*!
 * @file    fbdma.c
 * @brief   FBDMA driver
 *
 * Used to load data from FB to IMEM/DMEM and vice-versa.
 */

#include <stdint.h>
#include <stddef.h>
#include <misc/nvmisc_drf.h>

#include <fbdma/fbdma.h>
#include <cpu/io_local.h>

//
// MISRA 4.9 RFD: function-like macros
// MISRA 15.1 RFD: Forward gotos
// MISRA 8.7 RFD: Functions with external linkage only referenced in one translation unit
//
START_RFD_BLOCK(MISRA, DEVIATE, Directive_4_9, "Approval: Bug 200531995, DR: SWE-FSP-012-SWSADR.docx",
                MISRA, DEVIATE, Rule_15_1, "Approval: Bug 3176459, DR: SWE-FSP-028-SWSADR.docx",
                MISRA, DEVIATE, Rule_8_7, "Approval: Bug 200531999, DR: SWE-FSP-015-SWSADR.docx")

#if NVRISCV_FEATURE_SCP
//TODO: scp.h has not yet been ported to libSCP. When it is ported, update this to the proper path.
#include <cpu/scp.h>
#include <misc/ct-assert.h>

#ifndef SCP_REGISTER_SIZE
#error "SCP_REGISTER_SIZE not defined!"
#endif

CT_ASSERT(SCP_REGISTER_SIZE % DMA_BLOCK_SIZE_MIN == 0U,
    "Minimum SCPDMA transfer size is not aligned to minimum FBDMA block size!");
#endif

#if (NVRISCV_IMEM_SIZE == 0U) || (NVRISCV_DMEM_SIZE == 0U)
#error "NVRISCV_IMEM_SIZE and NVRISCV_DMEM_SIZE must be configured"
#endif

static error_t dmaWaitCompletion(void);

/**
 * @brief Create a NV_PRGNLCL_FALCON_DMATRFCMD register value for a dma transfer
 *
 * @param[in] tcmPa The RISCV PA for the TCM buffer. Must be aligned to DMA_BLOCK_SIZE_MIN,
 * but DMA_BLOCK_SIZE_MAX alignement is optimal.
 * @param[in] apertureOffset The external address (GPU PA or GPU VA) in the aperture
 * determined by dmaIdx. The type of memory to access is determined by FBIF_APERTURE_CFG.memType
 * (on engines with FBIF) or by TFBIF_APERTURE_CFG.swid (on engines with TFBIF).
 * riscvPaToTargetOffset can be used to translate a RISCV PA to a GPU PA/VA suitable for
 * this argument. Must be aligned to DMA_BLOCK_SIZE_MIN, but DMA_BLOCK_SIZE_MAX alignement
 * is optimal.
 * @param[in] sizeBytes The number of bytes to transfer. Must have a granularity of
 * DMA_BLOCK_SIZE_MIN, but DMA_BLOCK_SIZE_MAX granularity is optimal.
 * @param[in] dmaIdx Aperture index to use (0-7).
 * @param[in] bReadExt Transfer direction. true: External->TCM. false: TCM->External
 * @param[out] pDmaCmd The NV_PRGNLCL_FALCON_DMATRFCMD register value to use for this transfer.
 * @param[out] pTcmOffset The offset of tcmPa from the start of its TCM.
 *
 * @return E_SUCCESS on success.
 *         E_INVALID_PARAM if tcmPa is not a valid TCM address or if sizeBytes, tcmOffset,
 * or apertureOffset are not properly aligned.
 */
static error_t
createDmaCmd(
    uint64_t tcmPa,
    uint64_t sizeBytes,
    uint8_t  dmaIdx,
    bool     bReadExt,
    uint32_t *pDmaCmd,
    uint32_t *pTcmOffset
)
{
    error_t status = E_SUCCESS;
    bool bTcmImem = false;
    uint32_t dmaCmd = 0;
    uint32_t tcmOffset = 0;
    uint64_t tcmBufEnd;

    // overflow check for tcmPa + sizeBytes
    if ((UINT64_MAX - tcmPa) < sizeBytes)
    {
        status = E_INVALID_PARAM;
        goto out;
    }
    else
    {
        tcmBufEnd = tcmPa + sizeBytes;
    }

    // tcmPa is in ITCM or DTCM?
    if ((tcmPa >= NV_RISCV_AMAP_IMEM_START) &&
        (tcmBufEnd < (NV_RISCV_AMAP_IMEM_START + (uint64_t)NVRISCV_IMEM_SIZE)))
    {
        bTcmImem = true;
        tcmOffset = (uint32_t)((tcmPa - NV_RISCV_AMAP_IMEM_START) & 0xFFFFFFFFU);
    }
    else if ((tcmPa >= NV_RISCV_AMAP_DMEM_START) &&
        (tcmBufEnd < (NV_RISCV_AMAP_DMEM_START + (uint64_t)NVRISCV_DMEM_SIZE)))
    {
        tcmOffset = (uint32_t)((tcmPa - NV_RISCV_AMAP_DMEM_START) & 0xFFFFFFFFU);
    }
    else
    {
        status = E_INVALID_PARAM;
        goto out;
    }

    if (bTcmImem)
    {
        dmaCmd = FLD_SET_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _IMEM, _TRUE, dmaCmd);
    }
    else
    {
        dmaCmd = FLD_SET_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _IMEM, _FALSE, dmaCmd);
        dmaCmd = FLD_SET_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _SET_DMTAG, _TRUE, dmaCmd);
    }

    if (bReadExt) // Ext -> TCM
    {
        dmaCmd = FLD_SET_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _WRITE, _FALSE, dmaCmd);
    }
    else // TCM -> Ext
    {
        dmaCmd = FLD_SET_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _WRITE, _TRUE, dmaCmd);
    }

    if (dmaIdx <= DMA_MAX_DMAIDX)
    {
        dmaCmd = FLD_SET_DRF_NUM(_PRGNLCL, _FALCON_DMATRFCMD, _CTXDMA, dmaIdx, dmaCmd);
    }
    else
    {
        status = E_INVALID_PARAM;
        goto out;
    }

    *pDmaCmd = dmaCmd;
    *pTcmOffset = tcmOffset;

out:
    return status;
}

/**
 * @brief Kick off a DMA transfer between TCM PA and External memory
 *
 * @param[in] tcmOffset The offset of the TCM address from the start of that TCM.
 * @param[in] apertureOffset The external address (GPU PA or GPU VA) in the aperture
 *            determined by dmaIdx. This can be a GPU VA or GPU PA.
 * @param[in] sizeBytes The number of bytes to DMA.
 * @param[in] dmaCmd The value to write into the NV_PRGNLCL_FALCON_DMATRFCMD register,
 * excluding the size field. size will be determined inside this function.
 *
 * @return E_SUCCESS on success.
 *         E_INVALID_PARAM if apertureOffset is not a valid address or sizeBytes, tcmPa,
 *          or apertureOffset do not meet minimum alignment requirements.
 *         E_BUSY if there is a DMA NACK.
 *         E_FAULT if there is a transfer error.
 *
 * @pre The FBIF/TFBIF aperture referenced by dmaIdx must have been configured by
 * fbifConfigureAperture or tfbifConfigureAperture. If the aperture is configured
 * to use a GPU VA, a context must be bound.
 */
static error_t
dmaXfer(
    uint32_t tcmOffset,
    uint64_t apertureOffset,
    uint64_t sizeBytes,
    uint32_t dmaCmd
)
{
    error_t err = E_SUCCESS;
    uint32_t dmaBlockSize;
    uint8_t dmaEncBlockSize;
    uint32_t currentTcmOffset = tcmOffset;
    uint64_t bytesRemaining = sizeBytes;
    uint32_t dmaCmdWithBlockSize = dmaCmd;
    uint32_t fbOffset;

    if (((apertureOffset | sizeBytes | tcmOffset) & (DMA_BLOCK_SIZE_MIN - 1U)) != 0U)
    {
        err = E_INVALID_PARAM;
        goto out;
    }

    // When REQUIRE_CTX = true, DMAs without a CTX bound will report idle immediately but won't actually be issued
    localWrite(NV_PRGNLCL_FALCON_DMACTL,
        DRF_DEF(_PRGNLCL, _FALCON_DMACTL, _REQUIRE_CTX, _FALSE));

    //
    // Break up apertureOffset into a base/offset to be used in DMATRFBASE, TRFFBOFFS
    // Note: We cannot use currentTcmOffset for both TRFMOFFS and TRFFBOFFS because this
    // would require us to calculate a DMATRFBASE value by subtracting
    // (apertureOffset - currentTcmOffset) and this may result in a misaligned value
    // which we cannot program into DMATRFBASE.
    //
    fbOffset = (uint32_t) (apertureOffset & 0xFFU);

    // Check for overflows (CERT-C INT30 violation in currentTcmOffset += dmaBlockSize; fbOffset += dmaBlockSize;)
    if (((UINT64_MAX - currentTcmOffset) < sizeBytes) ||
        ((UINT64_MAX - fbOffset) < sizeBytes))
    {
        err = E_INVALID_PARAM;
        goto out;
    }

    localWrite(NV_PRGNLCL_FALCON_DMATRFBASE, (uint32_t)((apertureOffset >> 8U) & 0xffffffffU));
#ifdef NV_PRGNLCL_FALCON_DMATRFBASE1
    localWrite(NV_PRGNLCL_FALCON_DMATRFBASE1, (uint32_t)(((apertureOffset >> 8U) >> 32U) & 0xffffffffU));
#else //NV_PRGNLCL_FALCON_DMATRFBASE1
    if ((((apertureOffset >> 8U) >> 32U) & 0xffffffffU) != 0U)
    {
        err = E_INVALID_PARAM;
        goto out;
    }
#endif //NV_PRGNLCL_FALCON_DMATRFBASE1

    while (bytesRemaining != 0U)
    {
        // Wait if we're full
        while (FLD_TEST_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _FULL, _TRUE, localRead(NV_PRGNLCL_FALCON_DMATRFCMD)))
        {
        }

        localWrite(NV_PRGNLCL_FALCON_DMATRFMOFFS, currentTcmOffset); // This is fine as long as caller does error checking.
        localWrite(NV_PRGNLCL_FALCON_DMATRFFBOFFS, fbOffset);

        // Determine the largest pow2 block transfer size of the remaining data
        dmaBlockSize = LOWESTBIT(currentTcmOffset | fbOffset | DMA_BLOCK_SIZE_MAX);

        // Reduce the dmaBlockSize to not exceed remaining data.
        while (dmaBlockSize > bytesRemaining)
        {
            dmaBlockSize >>= 1U;
        }

        // Convert the pow2 block size to block size encoding
        dmaEncBlockSize = BIT_IDX_32(dmaBlockSize / DMA_BLOCK_SIZE_MIN);
        dmaCmdWithBlockSize = FLD_SET_DRF_NUM(_PRGNLCL, _FALCON_DMATRFCMD, _SIZE, dmaEncBlockSize, dmaCmd);

        localWrite(NV_PRGNLCL_FALCON_DMATRFCMD, dmaCmdWithBlockSize);

        bytesRemaining -= dmaBlockSize;
        currentTcmOffset += dmaBlockSize;
        fbOffset += dmaBlockSize;
    }

out:
    return err;
}

error_t
dmaPa
(
    uint64_t tcmPa,
    uint64_t apertureOffset,
    uint64_t sizeBytes,
    uint8_t  dmaIdx,
    bool     bReadExt
)
{
    error_t ret;
    uint32_t tcmOffset;
    uint32_t dmaCmd;

    ret = createDmaCmd(tcmPa, sizeBytes, dmaIdx, bReadExt, &dmaCmd, &tcmOffset);
    if (ret != E_SUCCESS)
    {
        goto out;
    }

    // Nothing to copy, bail after checks
    if (sizeBytes == 0U)
    {
        ret = E_SUCCESS;
        goto out;
    }

    ret = dmaXfer(tcmOffset, apertureOffset, sizeBytes, dmaCmd);
    if (ret != E_SUCCESS)
    {
        goto out;
    }

    INLINE_RFD(MISRA, DEVIATE, Directive_4_7, "Approval: JIRA TID-1795, DR: SWE-FSP-062-SWSADR.docx");
    ret = dmaWaitCompletion();

out:
    return ret;
}

/**
 * @brief Wait for all DMA transfers to complete
 *
 * @return E_SUCCESS on success.
 *         E_FAULT if DMA HW returned an error.
 *         E_BUSY if DMA engine has returned a NACK.
 *
 * @pre Should be called after dmaPaAsync to ensure that DMAs have completed.
 */
static error_t
dmaWaitCompletion(void)
{
    error_t err = E_SUCCESS;
    uint32_t dmatrfcmd;
    uint32_t dmainfo;

    // Wait for completion of any remaining transfers
    do {
        dmatrfcmd = localRead(NV_PRGNLCL_FALCON_DMATRFCMD);
    }
    while (FLD_TEST_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _IDLE, _FALSE, dmatrfcmd));

    if (FLD_TEST_DRF(_PRGNLCL, _FALCON_DMATRFCMD, _ERROR, _TRUE, dmatrfcmd))
    {
        err = E_FAULT;
        goto out;
    }

    dmainfo = localRead(NV_PRGNLCL_FALCON_DMAINFO_CTL);
    if (FLD_TEST_DRF(_PRGNLCL, _FALCON_DMAINFO_CTL, _INTR_ERR_COMPLETION, _TRUE, dmainfo))
    {
        // Received DMA NACK, clear it and return error
        dmainfo = FLD_SET_DRF(_PRGNLCL, _FALCON_DMAINFO_CTL, _INTR_ERR_COMPLETION, _CLR, dmainfo);
        localWrite(NV_PRGNLCL_FALCON_DMAINFO_CTL, dmainfo);
        err = E_BUSY;
        goto out;
    }

out:
    return err;
}

#if NVRISCV_FEATURE_SCP || defined(TEST_SCPTOEXTMEM)
error_t
dmaScpToExtmem(
    uint64_t apertureOffset,
    uint64_t sizeBytes,
    uint8_t dmaIdx
)
{
    error_t ret;

    // Nothing to copy
    if (sizeBytes == 0U)
    {
        return E_SUCCESS;
    }

    if ((((sizeBytes | apertureOffset) & (SCP_REGISTER_SIZE - 1U)) != 0U) ||
        (dmaIdx > DMA_MAX_DMAIDX))
    {
        return E_INVALID_PARAM;
    }

    //
    // Set dmaCmd:
    // - write to Extmem
    // - enable direct bypass shortcut path from SCP
    // - dmaIdx
    //
    uint32_t dmaCmd = DRF_DEF(_PRGNLCL, _FALCON_DMATRFCMD, _WRITE, _TRUE) |
                DRF_NUM(_PRGNLCL, _FALCON_DMATRFCMD, _SEC, 2U) |
                DRF_NUM(_PRGNLCL, _FALCON_DMATRFCMD, _CTXDMA, dmaIdx);

    //
    // call dmaXfer with tcmOffset=0, since this value won't get used anyways:
    // Data is sourced from SCP.
    //
    ret = dmaXfer(0U, apertureOffset, sizeBytes, dmaCmd);
    if (ret != E_SUCCESS)
    {
        return ret;
    }

    return dmaWaitCompletion();
}
#endif // NVRISCV_FEATURE_SCP

END_RFD_BLOCK(MISRA, DEVIATE, Directive_4_9, "Approval: Bug 200531995, DR: SWE-FSP-012-SWSADR.docx",
              MISRA, DEVIATE, Rule_15_1, "Approval: Bug 3176459, DR: SWE-FSP-028-SWSADR.docx",
              MISRA, DEVIATE, Rule_8_7, "Approval: Bug 200531999, DR: SWE-FSP-015-SWSADR.docx")
