/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2020 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

/*!
 * @file    tfbif.c
 * @brief   TFBIF driver
 *
 * Used to configure TFBIF apertures
 */

#include <stddef.h>
#include <misc/nvmisc_drf.h>

#include <fbdma/tfbif.h>
#include <cpu/io_local.h>

//
// MISRA 4.9 RFD: function-like macros
// MISRA 15.1 RFD: Forward gotos
// MISRA 8.7 RFD: Functions with external linkage only referenced in one translation unit
//
START_RFD_BLOCK(MISRA, DEVIATE, Directive_4_9, "Approval: Bug 200531995, DR: SWE-FSP-012-SWSADR.docx",
                MISRA, DEVIATE, Rule_15_1, "Approval: Bug 3176459, DR: SWE-FSP-028-SWSADR.docx",
                MISRA, DEVIATE, Rule_8_7, "Approval: Bug 200531999, DR: SWE-FSP-015-SWSADR.docx")

//The register manuals currently don't define these as indexed fields.
//Because of this, we need to define our own. We have asked the hardware
//Folks to provide indexed field definitions in the manuals, so we don't
//need to do this here.
#define NV_PRGNLCL_TFBIF_TRANSCFG_ATT_SWID(i)     ((i)*4u+1U):((i)*4U)
#define NV_PRGNLCL_TFBIF_TRANSCFG_ATT_SWID_XMSB(i)     ((i)*4U+1U)
#define NV_PRGNLCL_TFBIF_TRANSCFG_ATT_SWID_XLSB(i)     ((i)*4U)

#define NV_PRGNLCL_TFBIF_REGIONCFG_T_VPR(i)       ((i)*4U+3U):((i)*4U+3U)
#define NV_PRGNLCL_TFBIF_REGIONCFG_T_VPR_XMSB(i)       ((i)*4U+3U)
#define NV_PRGNLCL_TFBIF_REGIONCFG_T_VPR_XLSB(i)       ((i)*4U+3U)

#define NV_PRGNLCL_TFBIF_REGIONCFG1_T_APERT_ID(i) ((i)*8U+4U):((i)*8U)
#define NV_PRGNLCL_TFBIF_REGIONCFG1_T_APERT_ID_XMSB(i) ((i)*8U+4U)
#define NV_PRGNLCL_TFBIF_REGIONCFG1_T_APERT_ID_XLSB(i) ((i)*8U)

#define NV_PRGNLCL_TFBIF_REGIONCFG2_T_APERT_ID(i) (((i)-4U)*8U+4U):(((i)-4U)*8U)
#define NV_PRGNLCL_TFBIF_REGIONCFG2_T_APERT_ID_XMSB(i) (((i)-4U)*8U+4U)
#define NV_PRGNLCL_TFBIF_REGIONCFG2_T_APERT_ID_XLSB(i) (((i)-4U)*8U)

error_t
tfbifConfigureAperture
(
    const TFBIF_APERTURE_CFG *cfgs,
    uint8_t numCfgs
)
{
    error_t err = E_SUCCESS;
    uint8_t idx;
    uint32_t transCfg, regionCfg, regionCfg1, regionCfg2;

    if (cfgs == NULL)
    {
        err = E_INVALID_PARAM;
        goto out;
    }

    transCfg = localRead(NV_PRGNLCL_TFBIF_TRANSCFG);
    regionCfg = localRead(NV_PRGNLCL_TFBIF_REGIONCFG);
    regionCfg1 = localRead(NV_PRGNLCL_TFBIF_REGIONCFG1);
    regionCfg2 = localRead(NV_PRGNLCL_TFBIF_REGIONCFG2);

    for (idx = 0; idx < numCfgs; idx++)
    {
        if (cfgs[idx].apertureIdx >= TFBIF_NUM_APERTURES ||
            cfgs[idx].swid >= (1 << DRF_SIZE(NV_PRGNLCL_TFBIF_TRANSCFG_ATT0_SWID)) ||
            cfgs[idx].apertureId >= (1 << DRF_SIZE(NV_PRGNLCL_TFBIF_REGIONCFG1_T0_APERT_ID))
        )
        {
            err = E_INVALID_PARAM;
            goto out;
        }

        transCfg = FLD_IDX_SET_DRF_NUM(_PRGNLCL, _TFBIF_TRANSCFG, _ATT_SWID, cfgs[idx].apertureIdx, cfgs[idx].swid, transCfg);
        regionCfg = FLD_IDX_SET_DRF_NUM(_PRGNLCL, _TFBIF_REGIONCFG, _T_VPR, cfgs[idx].apertureIdx, cfgs[idx].vpr, regionCfg);
        if (cfgs[idx].apertureIdx <= 3)
        {
            regionCfg1 = FLD_IDX_SET_DRF_NUM(_PRGNLCL, _TFBIF_REGIONCFG1, _T_APERT_ID, cfgs[idx].apertureIdx, cfgs[idx].apertureId, regionCfg1);
        }
        else
        {
            regionCfg2 = FLD_IDX_SET_DRF_NUM(_PRGNLCL, _TFBIF_REGIONCFG2, _T_APERT_ID, cfgs[idx].apertureIdx - 4, cfgs[idx].apertureId, regionCfg2);
        }
    }

    localWrite(NV_PRGNLCL_TFBIF_TRANSCFG, transCfg);
    localWrite(NV_PRGNLCL_TFBIF_REGIONCFG, regionCfg);
    localWrite(NV_PRGNLCL_TFBIF_REGIONCFG1, regionCfg1);
    localWrite(NV_PRGNLCL_TFBIF_REGIONCFG2, regionCfg2);

out:
    return err;
}

//
// we use function like macros
// we also use forward jumping gotos for error handling (early returns)
//
END_RFD_BLOCK(MISRA, DEVIATE, Directive_4_9, "Approval: Bug 200531995, DR: SWE-FSP-012-SWSADR.docx",
                MISRA, DEVIATE, Rule_15_1, "Approval: Bug 3176459, DR: SWE-FSP-028-SWSADR.docx",
                MISRA, DEVIATE, Rule_8_7, "Approval: Bug 200531999, DR: SWE-FSP-015-SWSADR.docx")
