/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2021 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */
#ifndef FBDMA__MEMUTILS_H
#define FBDMA__MEMUTILS_H
#define FSP__FBDMA__MEMUTILS_H 1

#include <stddef.h>
#include <stdint.h>

#include <error/common-errors.h>

#if NVRISCV_HAS_FBIF
#include <fbdma/fbif.h>
#endif // NVRISCV_HAS_FBIF


/**
 * @brief Memory targets supported by riscvPaToTargetOffset()
 *
 * @typedef-title RISCV_MEM_TARGET
 */
typedef enum
{
    RISCV_MEM_TARGET_IMEM,
    RISCV_MEM_TARGET_DMEM,
    RISCV_MEM_TARGET_FBGPA,
    RISCV_MEM_TARGET_SYSGPA,
} RISCV_MEM_TARGET;


#if NVRISCV_HAS_FBIF

/**
 * @brief Converts a RISCV PA to an FBIF target/offset pair. Target is one of
 * the external memory types which Peregrine can access via FBIF (FB or SYSMEM)
 * and offset is the offset within the physical address range of that memory type.
 *
 * @param[in]  pa       The RISCV PA to convert.
 *
 * @param[out] pTarget  The target aperture (e.g. COHERENT_SYSTEM) that the PA
 *                      resides within.
 *
 * @param[out] pOffset  The offset of the PA within the target aperture.
 *
 * @return E_SUCCESS       if successful.
 * E_INVALID_PARAM if PA is not within the range of a valid memory aperture (SYSMEM or FB).
 */
error_t riscvPaToFbifAperture(uintptr_t pa, FBIF_TRANSCFG_TARGET *pTarget,
    uint64_t *pOffset);

#endif // NVRISCV_HAS_FBIF

/**
 * @brief Converts a RISCV PA to a global target/offset pair. Target is one of
 * the available memory types which Peregrine can access (IMEM, DMEM, FB, SYSMEM)
 * and offset is the offset within the physical address range of that memory type.
 *
 * @param[in]  pa       The RISCV PA to convert.
 *
 * @param[out] pTarget  The memory target (e.g. SYSMEM) that the PA resides
 *                      within (optional).
 *
 * @param[out] pOffset  The offset of the PA within the memory target
 *                      (optional).
 *
 * @return E_SUCCESS       if successful.
 *         E_INVALID_PARAM if PA does not fall inside any of the supported memory targets.
 */
error_t riscvPaToTargetOffset(uintptr_t pa, RISCV_MEM_TARGET *pTarget,
    uint64_t *pOffset);

#endif //FBDMA__MEMUTILS_H
